import datetime
import nest
import nest.topology as tp
import numpy as np

from . import serialize

__all__ = [
    'run',
]


def getNodes(collection, meta):
  if 'spatial' in meta:
    nodes = nest.GetNodes(collection)[0]
  else:
    nodes = collection
  return nodes


def log(message):
  # print(message)
  return (str(datetime.datetime.now()), 'server', message)


def run(data):
  logs = []

  logs.append(log('Get request'))
  simtime = data.get('time', 1000.0)
  kernel = data.get('kernel', {})
  models = data['models']
  collections = data['collections']
  connectomes = data['connectomes']
  records = []
  collections_obj = []

  logs.append(log('Reset kernel'))
  nest.ResetKernel()

  logs.append(log('Set seed in numpy random'))
  np.random.seed(int(data.get('random_seed', 0)))

  logs.append(log('Set kernel status'))
  local_num_threads = int(kernel.get('local_num_threads', 1))
  rng_seeds = np.random.randint(0, 1000, local_num_threads).tolist()
  resolution = float(kernel.get('resolution', 1.0))
  kernel_dict = {
      'local_num_threads': local_num_threads,
      'resolution': resolution,
      'rng_seeds': rng_seeds,
  }
  nest.SetKernelStatus(kernel_dict)
  data['kernel'] = kernel_dict

  logs.append(log('Collect all recordables for multimeter'))
  for idx, collection in enumerate(collections):
    model = models[collection['model']]
    if model['existing'] != 'multimeter':
      continue

    if 'record_from' in model['params']:
      continue

    recs = list(filter(lambda conn: conn['pre'] == idx, connectomes))
    if len(recs) == 0:
      continue

    recordable_models = []
    for conn in recs:
      recordable_model = models[collections[conn['post']]['model']]
      recordable_models.append(recordable_model['existing'])
    recordable_models_set = list(set(recordable_models))
    assert len(recordable_models_set) == 1

    recordables = nest.GetDefaults(recordable_models_set[0], 'recordables')
    model['params']['record_from'] = list(map(str, recordables))

  logs.append(log('Copy models'))
  for new, model in models.items():
    params_serialized = serialize.model_params(model['existing'], model['params'])
    nest.CopyModel(model['existing'], new, params_serialized)

  logs.append(log('Create collections'))
  for idx, collection in enumerate(collections):
    collections[idx]['idx'] = idx
    if 'spatial' in collection:
      specs = collection['spatial']
      specs['elements'] = collection['model']
      obj = tp.CreateLayer(serialize.layer(specs))
      if 'positions' in specs:
        positions = specs['positions']
      else:
        positions = tp.GetPosition(nest.GetNodes(obj)[0])
      positions = np.round(positions, decimals=2).astype(float)
      collections[idx]['spatial']['positions'] = positions.tolist()
      collections[idx]['n'] = positions.shape[0]
      collections[idx]['ndim'] = positions.ndim
      collections[idx]['global_ids'] = nest.GetNodes(obj)[0]
    else:
      n = int(collection.get('n', 1))
      obj = nest.Create(collection['model'], n)
      collections[idx]['global_ids'] = list(obj)
      if collection['element_type'] == 'recorder':
        model = models[collection['model']]
        record = {
            'recorder': {
                'global_ids': list(obj),
                'idx': idx,
                'model': model['existing']
            }
        }
        if 'record_from' in model['params']:
          record['record_from'] = model['params']['record_from']
        records.append(record)
    collections_obj.append(obj)

  logs.append(log('Connect collections'))
  for connectome in connectomes:
    pre = collections[connectome['pre']]
    post = collections[connectome['post']]
    pre_obj = collections_obj[connectome['pre']]
    post_obj = collections_obj[connectome['post']]
    if ('spatial' in pre) and ('spatial' in post):
      projections = connectome['projections']
      tp.ConnectLayers(pre_obj, post_obj, serialize.projections(projections))
    else:
      conn_spec = connectome.get('conn_spec', 'all_to_all')
      syn_spec = connectome.get('syn_spec', 'static_synapse')
      # NEST 2.18
      pre_nodes = getNodes(pre_obj, pre)
      post_nodes = getNodes(post_obj, post)
      nest.Connect(pre_nodes, post_nodes, serialize.conn(conn_spec), serialize.syn(syn_spec))
      # NEST 3
      # nest.Connect(pre_obj, post_obj, serialize.conn(conn_spec), serialize.syn(syn_spec))

  logs.append(log('Start simulation'))
  nest.Simulate(float(simtime))

  logs.append(log('End simulation'))
  data['kernel']['time'] = nest.GetKernelStatus('time')

  logs.append(log('Serialize recording data'))
  ndigits = int(-1 * np.log10(resolution))
  for idx, record in enumerate(records):
    records[idx]['idx'] = idx
    global_ids = []
    if record['recorder']['model'] == 'spike_detector':
      rec = 'post'
      neuron = 'pre'
    else:
      rec = 'pre'
      neuron = 'post'
  for connectome in connectomes:
    if connectome[rec] == record['recorder']['idx']:
      collection = collections[connectome[neuron]]
      global_ids.extend(collection['global_ids'])
    records[idx]['global_ids'] = global_ids
    recorder_obj = collections_obj[record['recorder']['idx']]
    events = serialize.events(recorder_obj, ndigits)
    records[idx]['events'] = events
    records[idx]['senders'] = list(set(events['senders']))
  data['records'] = records

  return {'data': data, 'logs': logs}



__all__ = [
    'flatten',
    'unflatten',
    'utilize',
]


def flatten(dd, sep='.'):
  items = []
  for key, value in dd.items():
    new_key = sep + key
    if isinstance(value, collections.MutableMapping):
      items.extend(flatten(value, new_key, sep=sep).items())
    else:
      items.append((new_key, value))
  return dict(items)


def unflatten(dd):
  new_dict = dict()
  for key, value in dd.items():
    parts = key.split(".")
    d = new_dict
    for part in parts[:-1]:
      if part not in d:
        d[part] = dict()
      d = d[part]
    d[parts[-1]] = value
  return new_dict


def utilize(dd, sep='.'):
  new_dict = {}
  for key, value in dd.items():
    if key.endswith(sep + 'value'):
      new_key = key[:-6]
      value_type = dd[key + sep + 'type']
      if value_type == 'float':
        new_value = float(value)
      elif val_type == 'int':
        new_value = int(value)
      else:
        new_value = value
      new_dict[new_key] = new_value
    elif not dd.endswith(sep + 'type'):
      new_dict[key] = value
  return new_dict



import array
import numpy as np

__all__ = [
    'serialize',
]


SLILiterals = [
    'element_type',
    'model',
    'record_from',
    'record_to',
    'recordables',
    'synapse_model',
    'type_id',
]

params_infinite = [
    'V_min',
    'alpha',
]


def serialize(data, toFixed=False):

  if type(data) in [array.array, np.ndarray]:
    data = data.tolist()

  if isinstance(data, list) or isinstance(data, tuple):
    data = [serialize(d) for d in data]
    data.sort()

  elif isinstance(data, dict):
    for key, value in data.items():

      if key in SLILiterals:
        if isinstance(value, tuple):
          data[key] = [d.name for d in data[key]]
        else:
          data[key] = value.name

      elif key == 'events':
        for ekey, event in value.items():
          if type(event) is np.ndarray:
            data[key][ekey] = event.tolist()

      elif toFixed:
        data[key] = str(data[key])

      elif type(value) is np.ndarray:
        data[key] = value.tolist()

      elif key in params_infinite:
        if np.isinf(value):
          data[key] = str(value)

      elif isinstance(value, tuple) and len(value) > 0:
        if isinstance(value[0], tuple) and len(value[0]) > 0:
          data[key] = [[j.tolist() for j in i] for i in value]

  return data



import json

blocks = [
    'state',
    'initial_values',
    'equations',
    'parameters',
    'internals',
    'input',
    'update'
]


def load_json_file(filename):
  with open(filename, 'r') as f:
    lines = f.readlines()
  data_json = ''.join(lines)
  data = json.loads(data_json)
  return data


def write_block(header, content):
  return header + ':\n' + content + '\nend\n'


def json_to_nestml_format(data):
  s = ''
  if 'doc' in data:
    s += '/*\n' + data['doc'] + '\n*/\n'
  s += 'neuron %s:\n' % data['neuron']
  for block in blocks:
    if block in data:
      s += write_block(block, data[block])
  s += 'output: %s\n' % data['output']
  s += 'end\n'
  return s


def write_file(model, text):
  with open(model + '.nestml', 'w') as f:
    f.write(text)


if __name__ == '__main__':
  model = 'iaf_psc_alpha_generated'
  data = load_json_file(model + '.json')
  data_nestml_format = json_to_nestml_format(data)
  write_file(model, data_nestml_format)



import nest
import nest.topology as tp
import numpy as np

__all__ = [
    'model_params',
    'parameter',
    'conn',
    'syn',
    'events',
    'layer',
    'projections',
]


def _paramify(params, param_defaults):
  _params = {}
  for pkey, pval in params.items():
    if pkey == 'model':
      _params[pkey] = pval
    elif isinstance(pval, dict):
      _params[pkey] = _paramify(pval, param_defaults[pkey])
    elif pkey in param_defaults:
      ptype = type(param_defaults[pkey])
      if ptype == np.ndarray:
        _params[pkey] = np.array(
            pval, dtype=param_defaults[pkey].dtype)
      else:
        _params[pkey] = ptype(pval)
  return _params


def model_params(model, params):
  if len(params) == 0:
    return {}
  param_defaults = nest.GetDefaults(model)
  return _paramify(params, param_defaults)


def parameter(specs):
  if isinstance(specs, dict):
    specs['specs'] = dict([(key, float(val)) for (key, val) in specs['specs'].items()])
    return tp.CreateParameter(**specs)
  else:
    return float(specs)


def conn(specs):
  if isinstance(specs, str):
    return specs
  if len(specs) == 0:
    return 'all_to_all'
  return specs


def syn(specs):
  if isinstance(specs, str):
    return specs
  if len(specs) == 0:
    return 'static_synapse'
  spec_defaults = nest.GetDefaults(specs.get('model', 'static_synapse'))
  return _paramify(specs, spec_defaults)


def events(recId, ndigits=0):
  events = {}
  for eventKey, eventVal in nest.GetStatus(recId, 'events')[0].items():
    events[eventKey] = eventVal.tolist()
  return events


def layer(specs):
  newSpecs = {'elements': specs['elements']}
  newSpecs['center'] = np.round(specs['center'], decimals=2).astype(float).tolist()
  newSpecs['extent'] = np.round(specs['extent'], decimals=2).astype(float).tolist()
  if 'positions' in specs:
    positions = np.round(specs['positions'], decimals=2).astype(float).tolist()
    newSpecs['positions'] = positions
  else:
    newSpecs['rows'] = specs['rows']
    newSpecs['columns'] = specs['columns']
  return newSpecs


def mask(masktype, specs):
  if 'rectangular' in masktype:
    newSpecs = {
        'lower_left': np.array(specs['lower_left'], dtype=float).tolist(),
        'upper_right': np.array(specs['upper_right'], dtype=float).tolist(),
        'azimuth_angle': float(specs['azimuth_angle']),
    }
  elif 'circular' in masktype:
    newSpecs = {
        'radius': float(specs['radius'])
    }
  elif 'doughnut' in masktype:
    newSpecs = {
        'inner_radius': float(specs['inner_radius']),
        'outer_radius': float(specs['outer_radius']),
    }
  elif 'elliptical' in masktype:
    newSpecs = {
        'major_axis': float(specs['major_axis']),
        'minor_axis': float(specs['minor_axis']),
        'azimuth_angle': float(specs['azimuth_angle']),
        # 'anchor': np.array(specs['anchor']).tolist(),
    }
  else:
    newSpecs = {}
  return tp.CreateMask(masktype=masktype, specs=newSpecs)


def projections(specs):
  newSpecs = {}
  newSpecs['connection_type'] = specs.get('connection_type', 'divergent')
  if 'kernel' in specs:
    newSpecs['kernel'] = parameter(specs['kernel'])
  if 'number_of_connections' in specs:
    newSpecs['number_of_connections'] = int(specs['number_of_connections'])
  if 'mask' in specs:
    newSpecs['mask'] = mask(specs['mask']['masktype'], specs['mask']['specs'])
  newSpecs['allow_autapses'] = bool(specs.get('allow_autapses', True))
  newSpecs['allow_multapses'] = bool(specs.get('allow_multapses', True))
  newSpecs['weights'] = parameter(specs.get('weights', 1.))
  newSpecs['delays'] = parameter(specs.get('delay', 1.))
  return newSpecs



from .serializer import serialize

__all__ = [
    'get_or_error',
]


def get_or_error(func):
  def func_wrapper(request, call, data, *args, **kwargs):

    try:
      data = func(request, call, data, *args, **kwargs)

      if 'data' not in data['response']:
        return data

      response = data['response']['data']
      if response is not None:
        data['response']['data'] = serialize(response, toFixed=False)
      data['response']['status'] = 'ok'

    except Exception as e:
      data['response']['msg'] = str(e)
      data['response']['status'] = 'error'

    return data

  return func_wrapper



__all__ = [
    'data_and_args',
]


def data_and_args(request, call=None):

  url = request.url_rule.rule.split('/')[1]
  data = {
      'request': {
          'url': url,
      },
      'response': {}
  }

  if call:
    data['request']['call'] = call

  args = []
  kwargs = {}

  if request.is_json:
    json = data['request']['json'] = request.get_json()
    if isinstance(json, list):
      args = json
    elif isinstance(json, dict):
      args = json.get('args', args)
      kwargs = json.get('kwargs', json)

  elif len(request.form) > 0:
    if 'args' in request.form:
      args = data['request']['form'] = request.form.getlist('args')
    else:
      kwargs = data['request']['form'] = request.form.to_dict()

  elif (len(request.args) > 0):
    if 'args' in request.args:
      args = data['request']['args'] = request.args.getlist('args')
    else:
      kwargs = data['request']['args'] = request.args.to_dict()

  return data, args, kwargs



import os
import optparse
import datetime
import inspect

import flask
from flask import Flask, request, jsonify
from flask_cors import CORS, cross_origin

import nest
import nest.topology as topo

from .api import initializer as api_init
from .api.client import api_client
from . import scripts

from . import __version__


app = Flask(__name__)
CORS(app)

nest_calls = dir(nest)
nest_calls = list(filter(lambda x: not x.startswith('_'), nest_calls))
nest_calls.sort()

topo_calls = dir(topo)
topo_calls = list(filter(lambda x: not x.startswith('_'), topo_calls))
topo_calls.sort()


# --------------------------
# General request
# --------------------------

@app.route('/', methods=['GET'])
def index():
  response = {
      'server': {
          'version': __version__,
          'git': {
              'ref': 'http://www.github.com/babsey/nest-server',
              'tag': 'v' + '.'.join(__version__.split('.')[:-1])
          }
      },
      'simulator': {
          'version': nest.version().split('-')[1],
      },
  }
  return jsonify(response)


# --------------------------
# RESTful API
# --------------------------

@app.route('/api/nest', methods=['GET'])
@cross_origin()
def router_nest():
  data, args, kwargs = api_init.data_and_args(request)
  response = api_client(request, nest_calls, data)
  return jsonify(response)


@app.route('/api/nest/<call>', methods=['GET', 'POST'])
@cross_origin()
def router_nest_call(call):
  data, args, kwargs = api_init.data_and_args(request, call)
  if call in nest_calls:
    call = getattr(nest, call)
    response = api_client(request, call, data, *args, **kwargs)
  else:
    data['response']['msg'] = 'The request cannot be called in NEST.'
    data['response']['status'] = 'error'
    response = data
  return jsonify(response)


@app.route('/api/topo', methods=['GET'])
@app.route('/api/nest_topology', methods=['GET'])
@cross_origin()
def router_topo():
  data, args, kwargs = api_init.data_and_args(request)
  response = api_client(request, topo_calls, data)
  return jsonify(response)


@app.route('/api/topo/<call>', methods=['GET', 'POST'])
@app.route('/api/nest_topology/<call>', methods=['GET', 'POST'])
@cross_origin()
def router_topo_call(call):
  data, args, kwargs = api_init.data_and_args(request, call)
  if call in topo_calls:
    call = getattr(topo, call)
    response = api_client(request, call, data, *args, **kwargs)
  else:
    data['response']['msg'] = 'The request cannot be called in NEST Topology.'
    data['response']['status'] = 'error'
    response = data
  return jsonify(response)


# --------------------------
# Scripts
# --------------------------

@app.route('/script/<filename>/<call>', methods=['POST', 'OPTIONS'])
@cross_origin()
def script(filename, call):
  # print(request.get_json())
  try:
    script = getattr(scripts, filename)
    func = getattr(script, call)
    response = func(request.get_json())
  except Exception as e:
    print(e)
    error = {
        'name': getattr(e, 'errorname'),
        'message': getattr(e, 'errormessage').split(':')[-1]
    }
    response = {'error': error}
  return jsonify(response)


@app.route('/source', methods=['GET'])
@cross_origin()
def inspect_files():
  try:
    source = inspect.getsource(scripts)
    response = {
        'source': source,
    }
  except Exception as e:
    print(e)
    error = {
        'name': getattr(e, 'errorname'),
        'message': getattr(e, 'errormessage').split(':')[-1]
    }
    response = {'error': error}
  return jsonify(response)


@app.route('/source/<filename>', methods=['GET'])
@cross_origin()
def inspect_script(filename):
  try:
    script = getattr(scripts, filename)
    source = inspect.getsource(script)
    response = {
        'source': source,
    }
  except Exception as e:
    print(e)
    error = {
        'name': getattr(e, 'errorname'),
        'message': getattr(e, 'errormessage').split(':')[-1]
    }
    response = {'error': error}
  return jsonify(response)


@app.route('/source/<filename>/<call>', methods=['GET'])
@cross_origin()
def inspect_func(filename, call):
  try:
    script = getattr(scripts, filename)
    func = getattr(script, call)
    source = inspect.getsource(func)
    response = {
        'source': source,
    }
  except Exception as e:
    print(e)
    error = {
        'name': getattr(e, 'errorname'),
        'message': getattr(e, 'errormessage').split(':')[-1]
    }
    response = {'error': error}
  return jsonify(response)


if __name__ == "__main__":
  parser = optparse.OptionParser("usage: python main.py [options]")
  parser.add_option("-H", "--host", dest="hostname",
                    default="127.0.0.1", type="string",
                    help="specify hostname to run on")
  parser.add_option("-p", "--port", dest="port", default=5000,
                    type="int", help="port to run on")
  (options, args) = parser.parse_args()
  app.run(host=options.hostname, port=options.port)



import inspect
import nest

from .decorator import get_or_error

__all__ = [
    'api_client',
]


@get_or_error
def api_client(request, call, data, *args, **kwargs):

    if callable(call):
        data['request']['call'] = call.__name__

        if str(kwargs.get('return_doc', 'false')) == 'true':
            response = call.__doc__
        if str(kwargs.get('return_source', 'false')) == 'true':
            response = inspect.getsource(call)
        else:
            if call.__name__ == 'SetKernelStatus':
                kernelStatus = nest.GetKernelStatus()
                for paramKey, paramVal in kwargs['params'].items():
                    kwargs['params'][paramKey] = type(kernelStatus[paramKey])(paramVal)
            elif call.__name__ == 'SetStatus':
                status = nest.GetStatus(kwargs['nodes'])
                for paramKey, paramVal in kwargs['params'].items():
                    kwargs['params'][paramKey] = type(status[paramKey])(paramVal)
            response = call(*args, **kwargs)
    else:
        response = call

    data['response']['data'] = response

    return data



import nest
import os
import subprocess

from pynestml.frontend.pynestml_frontend import to_nest, install_nest
from . import converter

__all__ = [
    'install',
]

root_path = '/tmp'
models_path = os.path.join(root_path, 'models')
build_path = os.path.join(root_path, 'models-build')
nest_install_dir = '/home/spreizer/opt/nest-simulator'


def bashCommand(command):
  process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
  output, error = process.communicate()
  return {'output': output, 'error': error}


def deleteModels():
  return bashCommand('rm -f %s/*' % models_path)


def createModelDir():
  return bashCommand('mkdir -p %s' % models_path)


def install(data):

  # save nestml models to files
  filename = os.path.join(models_path, data['neuron'])
  data_nestml_format = converter.json_to_nestml_format(data)
  converter.write_file(filename, data_nestml_format)

  to_nest(models_path, build_path, module_name='nestmlmodule')
  install_nest(build_path, nest_install_dir)

  return {'data': data}


createModelDir()



import setuptools
import sys

from nest_server import __version__

assert sys.version_info >= (3,), "Python 3 is required to run NEST Server"

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nest-server",
    version=__version__,
    description="A server for NEST Simulator with REST API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/babsey/nest-server",
    license="MIT License",
    packages=setuptools.find_packages(),
    scripts=["bin/nest-server"],
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.6",
    install_requires=["numpy", "flask", "flask-cors", "uwsgi", "nestml"],
)
